#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "StackOps.h"
// print eror message and abort program
static void error(const char* message)
{
    puts(message);
    abort();
}
// double stack memory, presenving contents
static void doubleStack(StackT *s)
{
    if (NULL == (s->memory = (StackElementT *)realloc(s->memory, 2 * sizeof(StackElementT) * s->size)))
    {
        error("doubleStack(): no memory");
    }
    s->size = 2 * s->size;
}
/** \brief creates an array-based stack of dynamic size
 *
 * \param initial stack size
 * \return pointer to created stack header
 *
 */
StackT *createStack(size_t size)
{
    // sanitize size
    if (size == 0)
        size = STACK_DEFAULT_SIZE;
    else if (size > STACK_MAX_SIZE)
        size = STACK_MAX_SIZE;
    // alocate memory for stack headercell
    StackT *s;
    if (NULL == (s = (StackT *)malloc(sizeof(StackT))))
        error("createStack(): No memory"); // no return from error; abort program
    // allocate memory for stack elements
    if (NULL == (s->memory = (StackElementT *)malloc(sizeof(StackElementT) * size)))
        error("createStack(): No memory"); // no return from error; abort program
    s->size = size;
    s->count = 0; // no elements stored, yet
    s->top = STACK_EMPTY;
    return s;
}
/** \brief Purges the stack and removes the stack from memory
 *
 * \param target stack
 * \return void
 *
 */
extern StackT *deleteStack(StackT *s)
{
    if (s == NULL) return s;
    free(s->memory);
    free(s);
    return NULL;
}
/** \brief pop an element from the stack
 *
 * \param target stack
 * \param result of operation (see StackStateT enum)
 * \return popped element
 *
 */
StackElementT pop(StackT *s, int *result)
{
    if (s == NULL)
    {
        *result = NO_STACK;
        return 0;
    }
    if (s->count == 0)
    {
        *result = STACK_EMPTY;
        return 0;
    }
    *result = STACK_OK;
    int oldTop = s->top;
    s->top--;
    if (s->top < 0) s->top = STACK_EMPTY;
    s->count--;
    return s->memory[oldTop];
}

/** \brief push an element onto given stack
 *
 * \param target stack
 * \param element value
 * \return result of operation (see StackStateT enum)
 *
 */
int push(StackT *s, StackElementT val)
{
    if (s == NULL)
    {
        return  NO_STACK;
    }
    if (s->top == s->size - 1)
    {
        // stack is full; attempt to double its size
        doubleStack(s);
    }
    s->top++;
    s->count++;
    s->memory[s->top] = val;
    return STACK_OK;
}

/** \brief peek at top element of the stack
 *
 * \param target stack
 * \param result of operation (see StackStateT enum)
 * \return value of top element
 *
 */
StackElementT top(StackT *s, int *result)
{
    if (s == NULL)
    {
        *result = NO_STACK;
        return 0;
    }
    if (s->count == 0)
    {
        *result = STACK_EMPTY;
        return 0;
    }
    *result = STACK_OK;
    return s->memory[s->top];
}

/**
* Implementation of stack tests
*/
/** \brief Test stack creation
 *
 * \param void
 * \return void
 *
 */
extern void testCreateStack()
{
    StackT *s = createStack(0);
    // check default size
    assert(s->size == STACK_DEFAULT_SIZE);
    assert(s->count == 0);
    assert(s->top == STACK_EMPTY);
    s = deleteStack(s);
    // cheack a valid size
    s = createStack(2);
    assert(s->size == 2);
    assert(s->count == 0);
    s = deleteStack(s);
    assert(s == NULL);
    // check size > maximum allowed
    s = createStack(STACK_MAX_SIZE + 1);
    assert(s->size == STACK_MAX_SIZE);
    assert(s->count == 0);
    s = deleteStack(s);
}
/** \brief Test stack pop operation
 *
 * \param void
 * \return void
 *
 */
extern void testPop()
{
    StackT *s = NULL;
    int result = 0;
    pop(s, &result);
    assert(result == NO_STACK);
    s = createStack(10);
    // pop from empty stack
    pop(s, &result);
    assert(result == STACK_EMPTY);
    push(s, 1);
    assert(s->memory[0] == 1);
    // stack: 1 topElement: 1
    StackElementT val = pop(s, &result);
    assert(val == 1);
    assert(result == STACK_OK);
    push(s, 2);
    assert(s->memory[0] == 2);
    // stack: 2 topElement: 2
    val = pop(s, &result);
    assert(val == 2);
    val = pop(s, &result);
    assert(result == STACK_EMPTY);
    // stack: empty
    val = pop(s, &result);
    assert(result == STACK_EMPTY);
    s = deleteStack(s);
    assert(s == NULL);
}
/** \brief Test stack push operation
 *
 * \param void
 * \return void
 *
 */
extern void testPush()
{
    StackT *s = NULL;
    int result = 0;
    push(s, 1);
    assert(result == NO_STACK);
    s = createStack(2);
    assert(s->size == 2);
    assert(s->count == 0);
    push(s, 1);
    assert(s->memory[0] == 1);
    push(s, 2);
    assert(s->memory[0] == 1 && s->memory[1] == 2);
    // stack: 1, 2 topElement: 2
    push(s, 3);
    // the stack should be doubled in size
    assert(s->size == 4);
    assert(s->count == 3);
    assert(s->memory[2] == 3 && s->memory[1] == 2  && s->memory[0] == 1);
    // stack: 1, 2, 3 topElement: 3
    deleteStack(s);
}
/** \brief Test stack top operation
 *
 * \param void
 * \return void
 *
 */
extern void testTop()
{
    StackT *s = NULL;
    int result = 0;
    StackElementT val = 0;
    push(s, 1);
    assert(result == NO_STACK);
    s = createStack(2);
    push(s, 1);
    // stack: 1; topElement: 1
    assert(s->memory[0] == 1);
    val = top(s, &result);
    assert(result == STACK_OK);
    assert(val == 1);
    push(s, 2);
    assert(s->memory[0] == 1 && s->memory[1] == 2);
    // stack: 1, 2; topElement: 2
    val = top(s, &result);
    assert(result == STACK_OK);
    assert(val == 2);
    push(s, 3);
    // stack: 1, 2, 3; topElement: 3
    assert(s->memory[0] == 1 && s->memory[1] == 2 && s->memory[2] == 3);
    assert(s->count == 3);
    assert(s->size == 4);
    pop(s, &result);
    assert(s->memory[0] == 1 && s->memory[1] == 2);
    // stack: 1, 2; topElement: 2
    val = top(s, &result);
    assert(s->memory[0] == 1 && s->memory[1] == 2);
    assert(result == STACK_OK);
    assert (val == 2);
    assert (s->count == 2);
    assert (s->size == 4);
    deleteStack(s);
}
